/* #####################################################################
 * #                        REFLEXIVE FRAMEWORK                        #
 * #                        ___________________                        #
 * #                                                                   #
 * # Reflexive Framework, Copyright (C) 2007 Andries Inze              #
 * #                                                                   #
 * # This library  is free software; you can redistribute it and/or    #
 * # modify it under the terms of the GNU Lesser General Public        #
 * # License as published by the Free Software Foundation; either      #
 * # version 2.1 of the License, or (at your option) any later version.#
 * #                                                                   #
 * # This library  is distributed in the hope that it will be useful,  #
 * # but WITHOUT ANY WARRANTY; without even the implied warranty of    #
 * # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     #
 * # GNU Lesser General Public License for more details.               #
 * #                                                                   #
 * # You should have received a copy of the GNU Lesser General Public  #
 * # License along with this program; if not, write to the Free        #
 * # Software Foundation, Inc., 51 Franklin Street, Fifth Floor,       #
 * # Boston, MA  02110-1301  USA                                       #
 * #                                                                   #
 * #####################################################################
 */
package org.reflexiveframework.proxy;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;



import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.MethodInterceptor;
import net.sf.cglib.proxy.MethodProxy;

/**
 * Reflexive provides a clean way to handle string literals, 
 * by dynamically resolving properties.
 * 
 * Create a control object by invoking <code>create</code>.
 * Only use with control objects. Usage with normal objects is unsupported.
 * 
 * @author Andries Inze
 * 
 */
public class Reflexive implements MethodInterceptor {

	/**
	 * The instance of the singleton.
	 * Making this an on-access loading, since double checked locking is broken.
	 */
	private static Reflexive instance = new Reflexive();

	/** Handles the mapping between method and the proxy object. */
	private HashMap methodToObject;

	/**
	 * The method name that was invoke.
	 */
	private List previousMethods;

	/**
	 * The cglib enchancer
	 */
	private Enhancer e;

	private Reflexive() {
		previousMethods = new LinkedList();
		e = new Enhancer();
		methodToObject = new HashMap();
	}

	private Object newInstance(Class clazz) {
		try {
			e.setSuperclass(clazz);
			e.setCallback(this);
			Object bean = Enhancer.create(clazz, getInstance());
			return bean;
		} catch (Throwable e) {
			throw new Error(e.getMessage());
		}
	}

	/**
	 * Creates a new control object.
	 * 
	 * @param clazz
	 * @return
	 */
	public static Object create(final Class clazz) {
		return getInstance().newInstance(clazz);
	}

	/**
	 * @return the instance of the singleton
	 */
	private final static Reflexive getInstance() {
		return instance;
	}

	/**
	 * Resolves the property from the method that was invoked. Usage:
	 * <code>Reflexive.resolvePropertyName(myObject.getMyProperty())</code>
	 * This would return <code>myProperty</code>
	 * 
	 * Only get- and is-methods are supported.
	 * 
	 * @param object
	 *            The returned object after invoking the created reflexive
	 *            object
	 * @return the resolved propertyName
	 */
	public static String resolvePropertyName(final Object object) {
		List previousMethods = getInstance().previousMethods;

		String propertyName = 
			ReflexiveUtils.convertMethodNameToPropertyName(
					((Method)previousMethods.get(previousMethods.size()-1))
					.getName());

		previousMethods.clear();

		return ReflexiveUtils.makeNestedPath(propertyName);
	}
	
	/**
	 * @param number the number to process
	 * @see resolvePropertyName(Object)
	 * @return the resolved propertyName
	 */
	public static String resolvePropertyName(final long number) {
		return resolvePropertyName(null);
	}
	
	/**
	 * @param number the number to process
	 * @see resolvePropertyName(Object)
	 * @return the resolved propertyName
	 */
	public static String resolvePropertyName(final double number) {
		return resolvePropertyName(null);
	}

	/**
	 * @param number the number to process
	 * @see resolvePropertyName(Object)
	 * @return the resolved propertyName
	 */
	public static String resolvePropertyName(final boolean number) {
		return resolvePropertyName(null);
	}
	
	/**
	 * Resolves the property from the method that was invoked. Usage:
	 * <code>Reflexive.resolveNestedPropertyName(myObject.getMyProperty().myNestedProperty())</code>
	 * This would return <code>myProperty.myNestedProperty</code>
	 * 
	 * Only get- and is-methods are supported.
	 * 
	 * @param object
	 *            The returned object after invoking the created reflexive
	 *            object
	 * @return the resolved propertyName
	 */
	public static String resolveNestedPropertyName(final Object object) {

		List previousMethods = getInstance().previousMethods;

		String[] propertyNames = new String[previousMethods.size()];

		for (int i = 0; i < previousMethods.size(); i++) {
			propertyNames[i] = ReflexiveUtils
					.convertMethodNameToPropertyName(
							((Method) previousMethods.get(i)).getName());
		}

		previousMethods.clear();

		return ReflexiveUtils.makeNestedPath(propertyNames);

	}
	
	/**
	 * @see resolveNestedPropertyName(Object)
	 * @param value the returned value after the invocation
	 * @return the nested propertyPath
	 */
	public static String resolveNestedPropertyName(final long value) {
		return resolveNestedPropertyName(null);
	}
	
	/**
	 * @see resolveNestedPropertyName(Object)
	 * @param value the returned value after the invocation
	 * @return the nested propertyPath
	 */
	public static String resolveNestedPropertyName(final double value) {
		return resolveNestedPropertyName(null);
	}
	
	/**
	 * @see resolveNestedPropertyName(Object)
	 * @param value the returned value after the invocation
	 * @return the nested propertyPath
	 */
	public static String resolveNestedPropertyName(final boolean value) {
		return resolveNestedPropertyName(null);
	}

	/**
	 * @see net.sf.cglib.proxy.MethodInterceptor #intercept(java.lang.Object,
	 *      java.lang.reflect.Method, java.lang.Object[],
	 *      net.sf.cglib.proxy.MethodProxy)
	 */
	public Object intercept(final Object object, final Method method,
			final Object[] args, final MethodProxy methodProxy)
			throws Throwable {

		if (method.getName().equals("hashCode")) {
			return Integer.valueOf(method.getReturnType().hashCode());
		} else if(method.getName().equals("equals")) {
			return Boolean.FALSE;
		}

		if (!method.getName().startsWith("get")
				&& !method.getName().startsWith("is")) {
			throw new RuntimeException("Only getters (or issers) are supported");
		}

		if (method.getReturnType().equals(Void.TYPE)) {
			throw new RuntimeException("Can't use proxy in void return types");
		}

		previousMethods.add(method);

		if (method.getReturnType().equals(int.class)
				|| method.getReturnType().equals(boolean.class)
				|| method.getReturnType().equals(short.class)
				|| method.getReturnType().equals(double.class)
				|| method.getReturnType().equals(char.class)
				|| method.getReturnType().equals(long.class)
				|| method.getReturnType().equals(float.class)
				|| method.getReturnType().equals(byte.class)) {

				return null;
		} else {
			//Returning the cached proxy
			if (!methodToObject.containsKey(method)) {
				methodToObject.put(method, buildMethodReturnProxy(method));
			}
			return methodToObject.get(method);
		}
	}

	private Object buildMethodReturnProxy(final Method method) {
		Object proxy = null;

		if (method.getReturnType().equals(String.class)) {
			proxy = ReflexiveUtils.createRandomString();
		} else if (method.getReturnType().equals(Boolean.class)) {
			proxy = new Boolean(true);
		} else if (method.getReturnType().equals(Integer.class)) {
			proxy = new Integer(1000);
		} else if (method.getReturnType().equals(Double.class)) {
			proxy = new Double(1000.0);
		} else if (method.getReturnType().equals(Long.class)) {
			proxy = new Long(1000L);
		} else if (method.getReturnType().equals(Float.class)) {
			proxy = new Float(1000L);
		} else if (method.getReturnType().equals(Character.class)) {
			proxy = new Character('x');
		} else if (method.getReturnType().equals(Short.class)) {
			proxy = new Short((short) 1);
		} else if (method.getReturnType().equals(Byte.class)) {
			proxy = new Byte((byte) 1);
		} else {
			// TODO: return null, with previousMethod, if proxy creation
			// fails.
			e.setSuperclass(method.getReturnType());
			e.setCallback(this);
			Object bean = Enhancer
					.create(method.getReturnType(), getInstance());
			return bean;
		}
		return proxy;
	}
}
